from __future__ import annotations
from typing import TYPE_CHECKING
import itertools
import functools
import numpy as np
from numbers import Number
from .params import Params
import operator

if TYPE_CHECKING:
    from .pbox_abc import Pbox
    from .dss import DempsterShafer
    from .dependency import Dependency


# * ---------------  Frechet ops --------------- *#


def frechet_op(x: Pbox, y: Pbox, op=operator.add):
    """Frechet operation on two pboxes

    note:
        this corresponds to the Frank, Nelson and Sklar Frechet bounds implementation
    """

    assert x.steps == y.steps, "Pboxes must have the same number of steps"

    n = x.steps

    nleft = np.empty(n)
    nright = np.empty(n)

    for i in range(0, n):
        j = np.arange(i, n)
        k = np.arange(n - 1, i - 1, -1)
        nright[i] = np.min(op(x.right[j], y.right[k]))
        jj = np.arange(0, i + 1)
        kk = np.arange(i, -1, -1)
        nleft[i] = np.max(op(x.left[jj], y.left[kk]))

    nleft.sort()
    nright.sort()

    return nleft, nright


def naive_frechet_op(x: Pbox, y: Pbox, op):
    """The real naive Frechet bounds implementation

    note:
        counterpart of `naivefrechetpbox` in `pba.r`

    args:
        x, y (Pbox): pboxes to be operated on
        op (function): arithemtic operator to be applied, e.g. operator.add

    return:
        Zu, Zd (tuple): left and right bounds of the pbox
        - Zu: lower bound of the pbox
        - Zd: upper bound of the pbox

    example:
        >>> from pyuncertainnumber import pba
        >>> a = pba.normal([4,5], 1)
        >>> b = pba.uniform([3,4], [7,8])
        >>> left_bound, right_bound = pba.naive_frechet_op(a, b, operator.add)
    """

    assert x.steps == y.steps, "Pboxes must have the same number of steps"

    n = x.steps
    # right
    c_r = vectorized_cartesian_op(x.right, y.right, op)
    c_r.sort()
    Zd = c_r[(n * n - n) : (n * n)]

    # left
    c = vectorized_cartesian_op(x.left, y.left, op)
    c.sort()
    Zu = c[:n]
    return Zu, Zd


def new_naive_frechet_op(
    x: Pbox,
    y: Pbox,
    op,
    n_sam=Params.steps,
):
    """this is a slow version"""
    from functools import partial
    from pyuncertainnumber import b2b, make_vec_interval

    # p_vars = [convert_pbox(v) for v in vars]

    n = n_sam  # number of samples to be taken

    assert x.steps == y.steps, "Pboxes must have the same number of steps"

    # if n < x.steps:
    #     # this change when there's specified dependency structure
    #     alpha = np.squeeze(qmc.LatinHypercube(d=1).random(n=n_sam))
    #     itvs = [v.alpha_cut(alpha) for v in [x, y]]
    # else:
    itvs = [v.discretise(n_sam) for v in [x, y]]

    # TODO add parallel logic herein
    b2b_f = partial(b2b, func=lambda x: op(x[0], x[1]), interval_strategy="direct")
    container = [b2b_f(_item) for _item in itertools.product(*itvs)]
    container = make_vec_interval(container)

    Zu = np.sort(container.lo)  # left
    Zd = np.sort(container.hi)  # right
    Zu = Zu[:n]  # take the first n elements
    Zd = Zd[(n * n - n) : (n * n)]  # take the last n elements
    return Zu, Zd


# def new_vectorised_naive_frechet_op(x, y, op):
#     """The new vectorised Frechet ops"""
#     assert x.steps == y.steps, "Pboxes must have the same number of steps"
#     n = x.steps
#     Xd = np.tile(x.right, n)
#     Xu = np.tile(x.left, n)
#     Yd = np.tile(y.right, n)
#     Yu = np.tile(y.left, n)

#     c1 = op(Xu, Yu)
#     c2 = op(Xu, Yd)
#     c3 = op(Xd, Yu)
#     c4 = op(Xd, Yd)
#     Zu = np.sort(np.minimum.reduce([c1, c2, c3, c4]))
#     Zd = np.sort(np.maximum.reduce([c1, c2, c3, c4]))

#     # I believe there should not be condensation per se, but rather the following boundings
#     Zu = Zu[:n]
#     Zd = Zd[(n * n - n) : (n * n)]
#     return Zu, Zd


def perfect_op(x: Pbox, y: Pbox, op=operator.add):
    """perfect operation on two pboxes

    note:
        defined for addition and multiplication. Different for subtraction and division.
    """
    nleft = op(x.left, y.left)
    nright = op(x.right, y.right)

    nleft.sort()
    nright.sort()
    return nleft, nright


def opposite_op(x: Pbox, y: Pbox, op=operator.add):
    """opposite operation on two pboxes

    note:
        defined for addition and multiplication. Different for subtraction and division.
    """
    nleft = op(x.left, np.flip(y.left))
    nright = op(x.right, np.flip(y.right))
    nleft.sort()
    nright.sort()
    return nleft, nright


def independent_op(x: Pbox, y: Pbox, op=operator.add):
    """independent operation on two pboxes

    note:
        defined for addition and multiplication. Different for subtraction and division.
    """
    c1 = vectorized_cartesian_op(x.left, y.left, op)
    c2 = vectorized_cartesian_op(x.left, y.right, op)
    c3 = vectorized_cartesian_op(x.right, y.left, op)
    c4 = vectorized_cartesian_op(x.right, y.right, op)

    nleft = np.sort(np.minimum.reduce([c1, c2, c3, c4]))
    nright = np.sort(np.maximum.reduce([c1, c2, c3, c4]))

    return nleft, nright


def copula_op(x: Pbox, y: Pbox, dependency: Dependency, op=operator.add) -> Pbox:
    """Bivariate operation on two pboxes with a given copula

    return:
        Pbox: resulting pbox after applying the operation with the copula
    """
    from .intervals import intervalise
    from pyuncertainnumber.pba.aggregation import stacking

    A = x.to_numpy()
    B = y.to_numpy()

    # Compute all pairwise interval sums
    C = op(A[:, None, :], B[None, :, :])  # shape (200, 200, 2)
    cc = intervalise(C)
    cc_ = cc.ravel()  # shape: (40000,)

    # Compute discrete copula mass matrix
    C_ = functools.partial(C_func, copula=dependency.copula)
    p = copula_mass(C_, x.p_values, y.p_values)  # shape (200, 200)
    p = p.ravel()  # shape: (40000,)
    return stacking(cc_, weights=p)


def copula_dss_op(
    x: DempsterShafer, y: DempsterShafer, dependency, op=operator.add
) -> Pbox:
    """Bivariate operation on two DSS with a given copula

    note:
        This is a temporary function until DSS is fully integrated with Pbox operations

    return:
        Pbox: resulting DSS after applying the operation with the copula
    """
    from .intervals import intervalise
    from pyuncertainnumber.pba.aggregation import stacking

    A = x.intervals.to_numpy()
    B = y.intervals.to_numpy()

    # Compute all pairwise interval sums
    C = op(A[:, None, :], B[None, :, :])  # shape (200, 200, 2)
    cc = intervalise(C)
    cc_ = cc.ravel()  # shape: (40000,)

    # Compute discrete copula mass matrix
    u = np.cumsum(x.masses)
    v = np.cumsum(y.masses)
    C_ = functools.partial(C_func, copula=dependency.copula)
    p = copula_mass(C_, u, v)  # shape (200, 200)
    p = p.ravel()  # shape: (40000,)
    return stacking(cc_, weights=p)


# backup
# def independent_op(x: Pbox, y: Pbox, op=operator.add):
#     """independent operation on two pboxes

#     note:
#         defined for addition and multiplication. Different for subtraction and division.
#     """
#     nleft = vectorized_cartesian_op(x.left, y.left, op)
#     nright = vectorized_cartesian_op(x.right, y.right, op)
#     return nleft, nright


def C_func(u, v, copula):
    # statsmodels expects columns; reshape back
    uv = np.column_stack((u.ravel(), v.ravel()))
    out = copula.cdf(uv)
    return out.reshape(u.shape)


def copula_mass(C_func, u, v, eps=1e-12):
    """
    Compute the discrete copula mass matrix p_ij from grids u, v on [0,1].
    C: callable taking arrays (same shape) and returning C(u,v)
    u, v: 1D arrays (ascending) of points in [0,1]
    eps: small value to avoid ±inf when calling C near 0 or 1
    """
    u = np.asarray(u, float)
    v = np.asarray(v, float)

    # extend with 0 for the finite-difference scheme
    u_ext = np.concatenate(([0.0], u))
    v_ext = np.concatenate(([0.0], v))

    # build grid
    U, V = np.meshgrid(u_ext, v_ext, indexing="ij")

    # safe evaluate: clip to (eps, 1-eps) for the C call
    Uc = np.clip(U, eps, 1.0 - eps)
    Vc = np.clip(V, eps, 1.0 - eps)
    Cgrid = C_func(Uc, Vc)

    # enforce copula axioms on boundaries exactly
    Cgrid[0, :] = 0.0  # C(0, v) = 0
    Cgrid[:, 0] = 0.0  # C(u, 0) = 0

    # rows/cols equal to 1.0, if present
    i1 = np.where(np.isclose(u_ext, 1.0))[0]
    j1 = np.where(np.isclose(v_ext, 1.0))[0]
    if i1.size:
        Cgrid[i1[0], :] = v_ext  # C(1, v) = v
    if j1.size:
        Cgrid[:, j1[0]] = u_ext  # C(u, 1) = u
    if i1.size and j1.size:
        Cgrid[i1[0], j1[0]] = 1.0  # C(1,1) = 1

    # 2D finite differences → copula mass
    p = Cgrid[1:, 1:] - Cgrid[:-1, 1:] - Cgrid[1:, :-1] + Cgrid[:-1, :-1]
    return p


def positiveconv_pbox(a, b, op=operator.add):
    """positive dependence (PQD) convolution of two pboxes

    example:
        >>> X = pba.uniform(1, 24)
        >>> Y = X
        >>> p_ = positiveconv_pbox(X, Y)
    """
    from .intervals import Interval
    from .pbox_abc import Staircase
    from pyuncertainnumber import envelope as env
    import math

    # positive (PQD) dependence
    assert a.steps == b.steps, "Pboxes must have the same number of steps"
    n = a.steps
    cu = np.zeros(n)
    cd = np.zeros(n)

    for i in range(n):
        infimum = np.inf
        for j in range(i, n):
            if op == operator.add:
                here = a.left[j] + b.left[int(n * i / (j + 1))]
            elif op == operator.mul:
                here = a.left[j] * b.left[int(n * i / (j + 1))]
            elif op == operator.pow:
                here = a.left[j] ** b.left[int(n * i / (j + 1))]
            if here < infimum:
                infimum = here
        cd[i] = infimum

        supremum = -np.inf
        for j in range(i + 1):
            kk = math.floor(1 + n * ((i) / n - j / n) / (1 - j / n)) - 1
            if op == operator.add:
                here = a.right[j] + b.right[kk]
            elif op == operator.mul:
                here = a.right[j] * b.right[kk]
            elif op == operator.pow:
                here = a.right[j] ** b.right[kk]
            if here > supremum:
                supremum = here
        cu[i] = supremum

    if op == "+":
        v = env(
            a.var + b.var,
            a.var + b.var + 2 * np.sqrt(a.var * b.var),
        ).to_interval()
    else:
        v = Interval(0, float("inf"))

    if op == "^" and a.range.hi <= 1:
        safe = cu
        cu = sorted(cd)
        cd = sorted(safe)

    return Staircase(
        cu,
        cd,
        mean=Interval(a.mean.lo + b.mean.lo, a.mean.hi + b.mean.hi),
        var=Interval(v.left, v.right),
    )


def new_vectorised_naive_frechet_op(x: Pbox, y: Pbox, op):
    """independent operation on two pboxes

    note:
        defined for addition and multiplication. Different for subtraction and division.
    """
    n = x.steps

    c1 = vectorized_cartesian_op(x.left, y.left, op)
    c2 = vectorized_cartesian_op(x.left, y.right, op)
    c3 = vectorized_cartesian_op(x.right, y.left, op)
    c4 = vectorized_cartesian_op(x.right, y.right, op)

    Zu = np.sort(np.minimum.reduce([c1, c2, c3, c4]))
    Zd = np.sort(np.maximum.reduce([c1, c2, c3, c4]))

    Zu = Zu[:n]
    Zd = Zd[(n * n - n) : (n * n)]
    return Zu, Zd
    # return c1, c2, c3, c4


def vectorized_cartesian_op(a, b, op):
    """vectorised cartesian operation for the bounds of pboxes

    note:
        used on self.left, other.left

    example:
        >>> nleft = vectorized_cartesian_op(self.left, other.left, operator.add)
        >>> nright = vectorized_cartesian_op(self.right, other.right, operator.add)
    """
    return op(a[:, np.newaxis], b).ravel()


def isum(l_p):
    """Sum of pboxes indepedently

    args:
        l_p (list): list of Pbox objects

    note:
        Same signature with Python ``sum`` which takes a list of inputs

    tip:
        Python ``sum`` accomplishes sum of Frechet case.
    """

    def binary_independent_sum(p1, p2):
        return p1.add(p2, dependency="i")

    return functools.reduce(binary_independent_sum, l_p)


# there is an new `convert` func
def convert(un):
    """Convert any input un into a Pbox object

    note:
        - theorically 'un' can be {Interval, DempsterShafer, Distribution, float, int}
    """

    from .pbox_abc import Pbox
    from .dss import DempsterShafer
    from .distributions import Distribution
    from .intervals.number import Interval

    if isinstance(un, Pbox):
        return un
    elif isinstance(un, Interval):
        return un.to_pbox()
    elif isinstance(un, Distribution):
        return un.to_pbox()
    elif isinstance(un, DempsterShafer):
        return un.to_pbox()
    elif isinstance(un, Number):
        return Interval(un, un).to_pbox()
    else:
        raise TypeError(f"Unable to convert {type(un)} object to Pbox")


def p_backcalc(a, c, ops):
    """backcal for p-boxes
    #! incorrect implementation
    args:
        a, c (Pbox):probability box objects
        ops (object) : {'additive_bcc', 'multiplicative_bcc'} whether additive or multiplicative
    """
    from pyuncertainnumber.pba.intervals.intervalOperators import make_vec_interval
    from pyuncertainnumber.pba.aggregation import stacking
    from .pbox_abc import Pbox
    from .intervals.number import Interval as I
    from .params import Params

    a_vs = a.to_interval()

    if isinstance(c, Pbox):
        c_vs = c.to_interval()
    elif isinstance(c, Number):
        c_vs = [I(c, c)] * Params.steps

    container = []
    for _item in itertools.product(a_vs, c_vs):
        container.append(ops(*_item))
    # print(len(container))  # shall be 40_000  # checkedout
    arr_interval = make_vec_interval(container)
    return stacking(arr_interval)


def adec(a, c):
    """
    Additive deconvolution: returns b such that a + b ≈ c
    Assumes a, b, c are instances of RandomNbr.

    note:
        implmentation from Scott
    """
    from .intervals.number import Interval as I
    from .pbox_abc import convert_pbox, Staircase

    n = Params.steps
    b = np.zeros(n)  # left bound of B, as in previous b.u[i]
    r = np.zeros(n)
    m = n - 1

    b[0] = c.left[0] - a.left[0]

    for i in range(1, m + 1):
        done = False
        sofar = c.left[i]
        for j in range(i):
            if sofar <= a.left[i - j] + b[j]:
                done = True
        if done:
            b[i] = b[i - 1]
        else:
            b[i] = c.left[i] - a.left[0]

    r[m] = c.right[m] - a.right[m]

    for i in range(m - 1, -1, -1):
        done = False
        sofar = c.right[i]
        for j in range(m, i, -1):
            if sofar >= a.right[i - j + m] + r[j]:
                done = True
        if done:
            r[i] = r[i + 1]
        else:
            r[i] = c.right[i] - a.right[m]

    # Check that bounds do not cross
    bad = any(b[i] > r[i] for i in range(n))

    if bad:
        # Try alternate method
        x = float("inf")
        y = float("-inf")
        for i in range(n):
            y = max(y, c.left[i] - a.left[i])
            x = min(x, c.right[i] - a.right[i])
        B = convert_pbox(I(y, x))
        return B

    # Final bounds check
    for i in range(n):
        if b[i] > r[i]:
            raise ValueError("Math Problem: couldn't deconvolve")
    return Staircase(left=b, right=r)


# * --------------- independent ops --------------- *#


def i_mul(a, b):
    """bivariate independent multiplication of pboxes"""

    if isinstance(a, Number) or isinstance(b, Number):
        return a * b
    else:  # 2 pboxes
        a, b = convert(a), convert(b)
        return a.mul(b, dependency="i")


# * --------------- vectorisation --------------- *#

###### base implementation of vector and matrix operations succeeds


class Vector:
    def __init__(self, components):
        self.components = components

    def __iter__(self):
        return iter(self.components)

    def __len__(self):
        return len(self.components)

    def __repr__(self):
        return f"Vector({self.components})"

    def __getitem__(self, index):
        if isinstance(index, slice):
            return Vector(self.components[index])
        else:
            return self.components[index]

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            if isinstance(value, Vector):
                self.components[index] = value.components
            elif isinstance(value, list):
                self.components[index] = value
            else:
                raise TypeError(
                    "Assigned value must be a Vector or list for slice assignment"
                )
        else:
            self.components[index] = value

    def __add__(self, other):
        if isinstance(other, Vector | list):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return Vector([a + b for a, b in zip(self, other)])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Vector):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return Vector([a - b for a, b in zip(self, other)])
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Vector([x * other for x in self.components])
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Vector([x / other for x in self.components])
        return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, Vector | list):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return sum(x * y for x, y in zip(self, other))

        elif isinstance(other, Matrix):
            # Vector @ Matrix: treat self as row vector, multiply by matrix
            n = len(self)
            m_rows, m_cols = other.shape()
            if n != m_rows:
                raise ValueError(
                    "Vector length must match matrix rows (row vector @ matrix)"
                )
            result = []
            for col in zip(*other.rows):
                result.append(sum(v * c for v, c in zip(self.components, col)))
            return Vector(result)

        return NotImplemented

    def tanh(self):
        """Pbox tanh operation elemenwise"""
        return Vector([x.tanh() for x in self.components])


class Matrix:
    def __init__(self, rows):
        # Handle case where input is a single Vector
        if isinstance(rows, Vector):
            # Turn Vector into a single-row Matrix
            self.rows = [list(rows.components)]
        elif isinstance(rows, list | np.ndarray):
            # Check if each row is a Vector, if so convert to list
            processed_rows = []
            for row in rows:
                if isinstance(row, Vector):
                    processed_rows.append(list(row.components))
                elif isinstance(row, list | np.ndarray):
                    processed_rows.append(row)
                else:
                    raise TypeError(f"Unsupported row type: {type(row)}")

            # Verify all rows have the same length
            if not all(len(row) == len(processed_rows[0]) for row in processed_rows):
                raise ValueError("All rows must have the same length")

            self.rows = processed_rows
        else:
            raise TypeError(f"Unsupported input type: {type(rows)}")

    def __getitem__(self, index):
        return self.rows[index]

    def __len__(self):
        return len(self.rows)

    def shape(self):
        return len(self.rows), len(self.rows[0])

    def __repr__(self):
        return f"Matrix({self.rows})"

    def __getitem__(self, index):
        if isinstance(index, tuple):
            # Expecting (row, col)
            row, col = index
            return self.rows[row][col]
        elif isinstance(index, slice):
            # Slice rows, return a new Matrix
            return Matrix(self.rows[index])
        else:
            # Single row, return it as Vector or list; here I use Vector for consistency
            return Vector(self.rows[index])

    def __setitem__(self, index, value):
        if isinstance(index, tuple):
            # index is (row, col)
            row, col = index
            self.rows[row][col] = value

        elif isinstance(index, slice):
            # index is a slice of rows
            if isinstance(value, Matrix):
                self.rows[index] = value.rows
            elif isinstance(value, list):
                self.rows[index] = value
            else:
                raise TypeError(
                    "Assigned value must be a Matrix or list of rows for slice assignment"
                )

        else:
            # Single row assignment
            if isinstance(value, Vector):
                self.rows[index] = value.components
            elif isinstance(value, list):
                self.rows[index] = value
            else:
                raise TypeError(
                    "Assigned value must be a Vector or list for row assignment"
                )

    def __add__(self, other):
        if isinstance(other, Matrix):
            if self.shape() != other.shape():
                raise ValueError("Matrices must have the same shape")
            return Matrix(
                [
                    [a + b for a, b in zip(row1, row2)]
                    for row1, row2 in zip(self.rows, other.rows)
                ]
            )
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, Matrix):
            if self.shape() != other.shape():
                raise ValueError("Matrices must have the same shape")
            return Matrix(
                [
                    [a - b for a, b in zip(row1, row2)]
                    for row1, row2 in zip(self.rows, other.rows)
                ]
            )
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Matrix([[x * other for x in row] for row in self.rows])
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Matrix([[x / other for x in row] for row in self.rows])
        return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, Vector):
            # Matrix @ Vector
            m, n = self.shape()
            if n != len(other):
                raise ValueError("Matrix columns must match vector length")
            return Vector(
                [sum(r[i] * other.components[i] for i in range(n)) for r in self.rows]
            )

        elif isinstance(other, Matrix):
            # Matrix @ Matrix
            m1, n1 = self.shape()
            m2, n2 = other.shape()
            if n1 != m2:
                raise ValueError("Incompatible shapes for matrix multiplication")
            result_rows = []
            for row in self.rows:
                result_row = []
                for col in zip(*other.rows):
                    result_row.append(sum(r * c for r, c in zip(row, col)))
                result_rows.append(result_row)
            return Matrix(result_rows)

        return NotImplemented

    def tanh(self):
        return Matrix([[x.tanh() for x in row] for row in self.rows])


# * --------------- independet Vector and Matrix for Pbox operations --------------- *#
class iVector(Vector):
    """Independent Vector of Pboxes"""

    def __init__(self, components):
        super().__init__(components)

    def __add__(self, other):
        if isinstance(other, iVector | list):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return iVector([isum(a, b) for a, b in zip(self, other)])
        return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, iVector | list):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return isum([i_mul(x, y) for x, y in zip(self, other)])

        elif isinstance(other, Matrix):
            # Vector @ Matrix: treat self as row vector, multiply by matrix
            n = len(self)
            m_rows, m_cols = other.shape()
            if n != m_rows:
                raise ValueError(
                    "Vector length must match matrix rows (row vector @ matrix)"
                )
            result = []
            for col in zip(*other.rows):
                result.append(isum(i_mul(v, c) for v, c in zip(self.components, col)))
            return Vector(result)

        return NotImplemented


class iMatrix(Matrix):
    """Independent Matrix of Pboxes"""

    def __init__(self, rows):
        super().__init__(rows)

    def __add__(self, other):
        if isinstance(other, iMatrix):
            if self.shape() != other.shape():
                raise ValueError("Matrices must have the same shape")
            return iMatrix(
                [
                    [isum(a, b) for a, b in zip(row1, row2)]
                    for row1, row2 in zip(self.rows, other.rows)
                ]
            )
        return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, iVector):
            # Matrix @ Vector
            m, n = self.shape()
            if n != len(other):
                raise ValueError("Matrix columns must match vector length")
            return iVector(
                [
                    isum(i_mul(r[i], other.components[i]) for i in range(n))
                    for r in self.rows
                ]
            )

        elif isinstance(other, iMatrix):
            # Matrix @ Matrix
            m1, n1 = self.shape()
            m2, n2 = other.shape()
            if n1 != m2:
                raise ValueError("Incompatible shapes for matrix multiplication")
            result_rows = []
            for row in self.rows:
                result_row = []
                for col in zip(*other.rows):
                    result_row.append(isum(i_mul(r, c) for r, c in zip(row, col)))
                result_rows.append(result_row)
            return iMatrix(result_rows)

        return NotImplemented


### use this later as `x.mul(y, dependency="i")` affects generality with other data types
class Vector_pbox(Vector):
    """Vector of Pboxes"""

    def __init__(self, components):
        super().__init__(components)

    def __matmul__(self, other):
        if isinstance(other, Vector | list):
            if len(self) != len(other):
                raise ValueError("Vectors must be the same length")
            return sum(x.mul(y, dependency="i") for x, y in zip(self, other))

        elif isinstance(other, Matrix):
            # Vector @ Matrix: treat self as row vector, multiply by matrix
            n = len(self)
            m_rows, m_cols = other.shape()
            if n != m_rows:
                raise ValueError(
                    "Vector length must match matrix rows (row vector @ matrix)"
                )
            result = []
            for col in zip(*other.rows):
                result.append(
                    sum(v.mul(c, dependency="i") for v, c in zip(self.components, col))
                )
            return Vector(result)

        return NotImplemented


#### backup ####
# class Matrix:
#     def __init__(self, rows):
#         if not all(len(row) == len(rows[0]) for row in rows):
#             raise ValueError("All rows must have the same length")
#         self.rows = rows

#     def __getitem__(self, index):
#         return self.rows[index]

#     def __len__(self):
#         return len(self.rows)

#     def shape(self):
#         return len(self.rows), len(self.rows[0])

#     def __repr__(self):
#         return f"Matrix({self.rows})"

#     def __add__(self, other):
#         if isinstance(other, Matrix):
#             if self.shape() != other.shape():
#                 raise ValueError("Matrices must have the same shape")
#             return Matrix(
#                 [
#                     [a + b for a, b in zip(row1, row2)]
#                     for row1, row2 in zip(self.rows, other.rows)
#                 ]
#             )
#         return NotImplemented

#     def __sub__(self, other):
#         if isinstance(other, Matrix):
#             if self.shape() != other.shape():
#                 raise ValueError("Matrices must have the same shape")
#             return Matrix(
#                 [
#                     [a - b for a, b in zip(row1, row2)]
#                     for row1, row2 in zip(self.rows, other.rows)
#                 ]
#             )
#         return NotImplemented

#     def __mul__(self, other):
#         if isinstance(other, (int, float)):
#             return Matrix([[x * other for x in row] for row in self.rows])
#         return NotImplemented

#     def __rmul__(self, other):
#         return self * other

#     def __truediv__(self, other):
#         if isinstance(other, (int, float)):
#             return Matrix([[x / other for x in row] for row in self.rows])
#         return NotImplemented

#     def __matmul__(self, other):
#         if isinstance(other, Vector):
#             # Matrix @ Vector
#             m, n = self.shape()
#             if n != len(other):
#                 raise ValueError("Matrix columns must match vector length")
#             return Vector(
#                 [sum(r[i] * other.components[i] for i in range(n)) for r in self.rows]
#             )

#         elif isinstance(other, Matrix):
#             # Matrix @ Matrix
#             m1, n1 = self.shape()
#             m2, n2 = other.shape()
#             if n1 != m2:
#                 raise ValueError("Incompatible shapes for matrix multiplication")
#             result_rows = []
#             for row in self.rows:
#                 result_row = []
#                 for col in zip(*other.rows):
#                     result_row.append(sum(r * c for r, c in zip(row, col)))
#                 result_rows.append(result_row)
#             return Matrix(result_rows)

#         return NotImplemented



import numpy as np
from typing import Callable
from scipy.optimize import minimize
from pyuncertainnumber.propagation.utils import Propagation_results

# TODO: test needed for this function alone and being used in EpistemicPropagation


def local_optimisation_method(
    x: np.ndarray,
    f: Callable,
    x0: np.ndarray = None,
    results: Propagation_results = None,
    tol_loc: np.ndarray = None,
    options_loc: dict = None,
    *,
    method_loc="Nelder-Mead",
) -> Propagation_results:  # Specify return type
    """
         Performs local optimization to find both the minimum and maximum values of a given function, within specified bounds.

     args:
         x (np.ndarray): A 2D NumPy array where each row represents an input variable and
                         the two columns define its lower and upper bounds (interval).
         f (Callable): The objective function to be optimized. It should take a 1D NumPy array
                       as input and return a scalar value.
         results (Propagation_results, optional): An object to store propagation results.
                                             Defaults to None, in which case a new
                                             `Propagation_results` object is created.
         x0 (np.ndarray, optional): A 1D or 2D NumPy array representing the initial guess for the
                                     optimization.
                                     - If x0 has shape (n,), the same initial values are used for both
                                       minimization and maximization.
                                     - If x0 has shape (2, n), x0[0, :] is used for minimization and
                                       x0[1, :] for maximization.
                                     If not provided, the midpoint of each variable's interval is used.
                                     Defaults to None.
         tol_loc (np.ndarray, optional): Tolerance for termination.
                                         - If tol_loc is a scalar, the same tolerance is used for both
                                           minimization and maximization.
                                         - If tol_loc is an array of shape (2,), tol_loc[0] is used for
                                           minimization and tol_loc[1] for maximization.
                                         Defaults to None.
         options_loc (dict, optional): A dictionary of solver options.
                                       - If options_loc is a dictionary, the same options are used for
                                         both minimization and maximization.
                                       - If options_loc is a list of two dictionaries, options_loc[0]
                                         is used for minimization and options_loc[1] for maximization.
                                       Refer to `scipy.optimize.minimize` documentation for available
                                       options. Defaults to None.
         method_loc (str, optional): The optimization method to use (e.g., 'Nelder-Mead', 'COBYLA').
                                     Defaults to 'Nelder-Mead'.

    signature:
         local_optimisation_method(x:np.ndarray, f:Callable,  results = None,
                               *, x0:np.ndarray = None,
                               tol_loc:np.ndarray = None, options_loc: dict = None, method_loc = 'Nelder-Mead') -> dict

     notes:
         This function utilizes the `scipy.optimize.minimize` function to perform local optimization.
         Refer to `scipy.optimize.minimize` documentation for available options.
         It only handles a function which produces a single output.

     returns:
         An `Propagation_results` object which contains:
             - 'un': UncertainNumber object(s) to represent the interval of the output.
             - 'raw_data' (dict): Dictionary containing raw data shared across output:
                     - 'x' (np.ndarray): Input values.
                     - 'f' (np.ndarray): Output values.
                     - 'min' (np.ndarray): Array of dictionaries for the function's output,
                               containing 'f' for the minimum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                     - 'max' (np.ndarray): Array of dictionaries for the function's output,
                               containing 'f' for the maximum of that output as well 'message', 'nit', 'nfev', 'final_simplex'.
                     - 'bounds' (np.ndarray): 2D array of lower and upper bounds for the output.

     example:
         >>> f = lambda x: x[0] + x[1] + x[2]  # Example function
         >>> x_bounds = np.array([[1, 2], [3, 4], [5, 6]])
         >>> # Initial guess (same for min and max)
         >>> x0 = np.array([1.5, 3.5, 5.5])
         >>> # Different tolerances for min and max
         >>> tol_loc = np.array([1e-4, 1e-6])
         >>> # Different options for min and max
         >>> options_loc = [
         >>>     {'maxiter': 100},  # Options for minimization
         >>>     {'maxiter': 1000}  # Options for maximization
         >>>     ]
         >>> # Perform optimization
         >>> y = local_optimisation_method(x_bounds, f, x0=x0, tol_loc=tol_loc,
         >>>                                     options_loc=options_loc)

    """

    bounds = [(var[0], var[1]) for var in x]

    def negated_f(x):
        return -f(x)

    if x0 is None:
        x0 = np.mean(x, axis=1)  # Use midpoint of intervals as initial guess
    x0 = np.atleast_2d(x0)  # Ensure x0 is 2D
    if (
        x0.shape[0] == 1
    ):  # If only one initial guess is provided, use it for both min and max
        x0 = np.tile(x0, (2, 1))

    # Handle tol_loc
    if tol_loc is None:
        tol_min = None  # Use default tolerance
        tol_max = None  # Use default tolerance
    elif np.isscalar(tol_loc):  # If tol_loc is a scalar
        tol_min = tol_loc
        tol_max = tol_loc
    else:  # If tol_loc is an array
        tol_min = tol_loc[0]
        tol_max = tol_loc[1]

    #  Handle options_loc
    if options_loc is None:
        options_min = None  # Use default options
        options_max = None  # Use default options
    elif isinstance(options_loc, dict):  # If options_loc is a single dictionary
        options_min = options_loc
        options_max = options_loc
    else:  # If options_loc is a list of two dictionaries
        options_min = options_loc[0]
        options_max = options_loc[1]

    # Perform minimization and maximization
    min_y = minimize(
        f,
        x0=x0[0, :],
        method=method_loc,
        bounds=bounds,
        tol=tol_min,
        options=options_min,
    )
    max_y = minimize(
        negated_f,
        x0=x0[1, :],
        method=method_loc,
        bounds=bounds,
        tol=tol_max,
        options=options_max,
    )
    max_y.fun = -max_y.fun  # Correct the sign of the maximum value

    # Create an instance of the result class
    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Store ALL results in the results object with descriptions
    results.raw_data["min"] = np.append(
        results.raw_data["min"],
        {
            "x": min_y.x,
            "f": min_y.fun,
            "message": min_y.message,
            "niterations": min_y.nit,
            "nfevaluations": min_y.nfev,
            "final_simplex": min_y.final_simplex,
        },
    )

    results.raw_data["max"] = np.append(
        results.raw_data["max"],
        {
            "x": max_y.x,
            "f": max_y.fun,
            "message": max_y.message,
            "niterations": max_y.nit,
            "nfevaluations": max_y.nfev,
            "final_simplex": max_y.final_simplex,
        },
    )

    results.raw_data["bounds"] = np.array([min_y.fun, max_y.fun])

    return results


# # Example function
# f = lambda x: x[0] + x[1] + x[2]  # Example function
# x_bounds = np.array([[1, 2], [3, 4], [5, 6]])

# # Initial guess (same for min and max)
# x0 = np.array([1.5, 3.5, 5.5])

# # Different tolerances for min and max
# tol_loc = np.array([1e-4, 1e-6])

# # Different options for min and max
# options_loc = [
#     {'maxiter': 100},  # Options for minimization
#     {'maxiter': 1000}  # Options for maximization
#      ]

# # Perform optimization
# y = local_optimisation_method(x_bounds, f, x0=x0, tol_loc=tol_loc,
#                                 options_loc=options_loc)

# # Print the results
# y.print()



from __future__ import annotations
from typing import TYPE_CHECKING

from abc import ABC, abstractmethod
import numpy as np
from numpy.typing import ArrayLike
import scipy.stats as sps
from numbers import Number
from pyuncertainnumber.pba.pbox_abc import Pbox, Staircase
from bisect import bisect_left


class Joint(ABC):

    def __init__(self, copula, marginals: list):
        self.copula = copula
        self.marginals = marginals


def wasserstein_1d(q1: ArrayLike, q2: ArrayLike, p: ArrayLike) -> float:
    """An intuitive of Wasserstein metric in 1D, aka. area between two quantile functions

    This is equivaluent to the Area Metric in 1D, which shall return same results as "scipy.stats.wasserstein_distance"

    args:
        q1, q2 (ArrayLike): quantile vectors (same length, corresponding to probabilities p)

        p      (ArrayLike): probability vector (between 0 and 1, monotone increasing)
    """

    diff = np.abs(q1 - q2)
    return np.trapz(y=diff, x=p)


def area_metric_ecdf(q1, q2, p):
    """Wasserstein metric in 1D, aka. area between two quantile functions

    This is equivaluent to the Area Metric in 1D.

    args:
        q1, q2 (ArrayLike): quantile vectors (same length, corresponding to probabilities p)

        p      (ArrayLike): probability vector (between 0 and 1, monotone increasing).
                            Must be the same for q1 and q2
    """
    p = np.asarray(p)
    q1 = np.asarray(q1)
    q2 = np.asarray(q2)

    diff = np.abs(q1 - q2)  # broadcasts if q1 or q2 is scalar
    if diff.shape != p.shape:
        # allow (scalar) -> expand to match p
        if diff.ndim == 0:
            diff = np.full_like(p, diff, dtype=float)
        else:
            raise ValueError("q1 and q2 must be broadcastable to the shape of p.")
    return np.trapz(y=diff, x=p)


def endpoint_distance(A, B):
    """
    Smallest endpoint distance elementwise between intervals.
    """
    A = np.atleast_2d(A)
    B = np.atleast_2d(B)

    if A.shape != B.shape:
        raise ValueError(
            "For elementwise comparison, A and B must have the same shape."
        )

    # compute all 4 endpoint differences per pair
    diffs = np.abs(A[:, :, None] - B[:, None, :])  # shape (n,2,2)
    distances = diffs.min(axis=(1, 2))  # shape (n,)

    if len(distances) == 1:
        return distances.item()
    return distances


def am_distance_counter(A, B):
    """Compute the distance between two sets of intervals as used in the area metric.

    notes:
        It is essentially doing:

        def f(a, b, c, d):
            return np.maximum.reduce([c - b, a - d, 0])

    """
    a, b = A[:, 0], A[:, 1]
    c, d = B[:, 0], B[:, 1]
    return np.maximum.reduce([c - b, a - d, np.zeros_like(a)])


def function_succeeds(f, *args, **kwargs):
    try:
        f(*args, **kwargs)
        return True
    except Exception:
        return False


def area_metric_pbox(a: Pbox, b: Pbox):
    """when a and b are both Pboxes"""
    # diff = endpoint_distance(a.to_numpy(), b.to_numpy())  # old version busted by Scott
    diff = am_distance_counter(a.to_numpy(), b.to_numpy())
    return np.trapz(y=diff, x=a.p_values)


def area_metric_pbox_diff(a: Pbox, b: Pbox):
    """when a and b are both Pboxes"""
    # diff = endpoint_distance(a.to_numpy(), b.to_numpy())  # old version busted by Scott
    diff = am_distance_counter(a.to_numpy(), b.to_numpy())
    return diff


def area_metric_sample(a: ArrayLike, b: ArrayLike):
    return sps.wasserstein_distance(a, b)


def area_metric_number(a: Pbox | Number, b: Pbox | Number) -> float:
    """if any of a or b is a number, compute area metric accordingly"""
    from pyuncertainnumber import pba

    if isinstance(a, Number) and isinstance(b, Number):
        return abs(a - b)
    if isinstance(a, Number):
        a, b = b, a  # swap so b is the number
    if isinstance(a, Pbox) and a.degenerate:
        return area_metric_ecdf(a.left, b, a.p_values)
    if isinstance(a, Pbox) and (not a.degenerate):
        # make b a Pbox
        b = pba.I(b).to_pbox()
        return area_metric_pbox(a, b)


def area_metric(a: Number | Pbox | ArrayLike, b: Number | Pbox | ArrayLike) -> float:
    """Compute the area metric between two objects.

    note:
        top-level function to compute area metric between any two objects
    """
    if isinstance(a, Number) or isinstance(b, Number):
        return area_metric_number(a, b)
    if isinstance(a, Pbox) and isinstance(b, Pbox):
        if a.degenerate and b.degenerate:
            return area_metric_ecdf(a.left, b.left, a.p_values)
        elif function_succeeds(a.imp, b):
            return 0.0
        else:
            return area_metric_pbox(a, b)
    if isinstance(a, (np.ndarray, list)) and isinstance(b, (np.ndarray, list)):
        return area_metric_sample(a, b)
    elif (isinstance(a, Pbox) and isinstance(b, np.ndarray)) or (
        isinstance(a, np.ndarray) and isinstance(b, Pbox)
    ):
        # make a a Pbox and b a sample anyway
        if not isinstance(a, Pbox):
            a, b = b, a

        # b has to be a scalar sample
        b = np.squeeze(b).item()
        return area_metric_number(a, b)
    else:
        raise NotImplementedError("Area metric not implemented for these types.")


# * --------------------------- area metric hints plot


def am_diff_register(A, B, debug=False):
    """Register the distance and return a structured array with fields:

    returns:
        In each tuple of the output array:
        - distance: float (max of c-b, a-d, 0)
        - x1: first endpoint (c or a, or 0 if 0 wins)
        - x2: second endpoint (b or d, or 0 if 0 wins)
    """
    A = np.asarray(A)
    B = np.asarray(B)

    # Expect shape (N, 2): [:,0]=start=a/c, [:,1]=end=b/d
    a, b = A[:, 0], A[:, 1]
    c, d = B[:, 0], B[:, 1]

    c_minus_b = c - b
    a_minus_d = a - d
    zeros = np.zeros_like(a, dtype=np.result_type(a, b, c, d, float))

    # Stack and argmax exactly mirrors "max(c-b, a-d, 0)" with deterministic tie-breaking.
    stacked = np.stack([c_minus_b, a_minus_d, zeros], axis=1)
    max_indices = np.argmax(stacked, axis=1)
    distances = stacked[np.arange(len(a)), max_indices]

    # Structured result
    result = np.zeros(len(a), dtype=[("distance", "f8"), ("x1", "f8"), ("x2", "f8")])
    result["distance"] = distances

    mask_cb = max_indices == 0
    mask_ad = max_indices == 1
    mask_0 = max_indices == 2

    result["x1"][mask_cb] = c[mask_cb]
    result["x2"][mask_cb] = b[mask_cb]

    result["x1"][mask_ad] = a[mask_ad]
    result["x2"][mask_ad] = d[mask_ad]

    result["x1"][mask_0] = 0.0
    result["x2"][mask_0] = 0.0

    if debug:
        print("a:", a)
        print("b:", b)
        print("c:", c)
        print("d:", d)
        print("c-b:", c_minus_b)
        print("a-d:", a_minus_d)
        print("stacked:\n", stacked)
        print("argmax:", max_indices)
        print("distances:", distances)
        print("result:", result)

    return result


import numpy as np
import matplotlib.pyplot as plt


def intervals_from_res(res):
    """
    From a structured array `res` with fields: 'distance', 'x1', 'x2',
    return:
      - mask: boolean mask where distance != 0
      - intervals: (N, 2) array of sorted [lo, hi] endpoints for rows where mask is True
    """
    mask = res["distance"] != 0
    x1 = res["x1"][mask]
    x2 = res["x2"][mask]
    lo = np.minimum(x1, x2)
    hi = np.maximum(x1, x2)
    intervals = np.stack([lo, hi], axis=1)
    return mask, intervals


import numpy as np
import matplotlib.pyplot as plt


def plot_intervals_from_res(
    res, p, *, show_points=False, title=None, include_ties=False, ax=None
):
    """Plot horizontal intervals from `res` at heights given by `p`.

    args
        res (ArrayLike) : Structured array with fields ('distance', 'x1', 'x2').

        p (ArrayLike) : 1D array of same length as `res`, giving y-coordinate (height) of each interval.

        show_points (Boolean): If True, shows markers at interval endpoints.

        title (str): Plot title. Optional

        include_ties (bool): If True, also include intervals where distance == 0 but argmax picked c-b or a-d.

        ax (matplotlib.axes.Axes) :  Axis to plot on. If None, a new figure and axis are created. Optional

    returns
        ax (matplotlib.axes.Axes): The matplotlib axis used for plotting.
    """
    res = np.asarray(res)
    p = np.asarray(p)
    if res.shape[0] != p.shape[0]:
        raise ValueError("p must have the same number of rows as res")

    # Select rows to plot
    if include_ties:
        mask = ~((res["x1"] == 0) & (res["x2"] == 0))
    else:
        mask = res["distance"] > 0

    if ax is None:
        fig, ax = plt.subplots()

    if not np.any(mask):
        ax.set_xlabel("x")
        ax.set_ylabel("p (height)")
        if title:
            ax.set_title(title)
        return ax

    x1 = res["x1"][mask]
    x2 = res["x2"][mask]
    y = p[mask]

    lo = np.minimum(x1, x2)
    hi = np.maximum(x1, x2)

    # Draw intervals
    for xi0, xi1, yi in zip(lo, hi, y):
        ax.plot([xi0, xi1], [yi, yi], color="lavender")  # horizontal segment
        if show_points:
            ax.plot([xi0, xi1], [yi, yi], "o", color="C0")

    ax.set_xlabel("x")
    ax.set_ylabel("p (height)")
    if title:
        ax.set_title(title)

    # Set limits dynamically based on data
    ax.set_xlim(lo.min() - 0.1, hi.max() + 0.1)
    ax.set_ylim(y.min() - 0.1, y.max() + 0.1)

    return ax


import numpy as np


def integrate_distance(res, p):
    """
    Integrate distance vs p with the trapezoidal rule,
    matching: np.trapz(y=diff, x=p)

    No masking; uses all rows. Sorts by p to avoid signed/ordering issues.
    """
    res = np.asarray(res)
    p = np.asarray(p)
    if res.shape[0] != p.shape[0]:
        raise ValueError("`p` must have the same length as `res`.")

    order = np.argsort(p)
    y = res["distance"][order].astype(float, copy=False)
    x = p[order].astype(float, copy=False)

    return np.trapz(y=y, x=x)


# def plot_intervals_from_res(res, p, *, show_points=False, title=None, ax=None):
#     """
#     For each row where res['distance'] != 0, plot a horizontal segment from
#     min(x1, x2) to max(x1, x2) at height p[i].

#     Args
#     ----
#     res : structured np.ndarray with fields ('distance','x1','x2')
#     p   : 1D array-like of same length as res
#     show_points : if True, also plot small markers at the endpoints
#     title : optional figure title
#     """
#     res = np.asarray(res)
#     p = np.asarray(p)
#     if res.shape[0] != p.shape[0]:
#         raise ValueError("p must have the same number of rows as res")

#     mask, intervals = intervals_from_res(res)
#     if not np.any(mask):
#         # nothing to plot
#         fig, ax = plt.subplots()
#         ax.set_xlabel("x")
#         ax.set_ylabel("p")
#         if title:
#             ax.set_title(title)
#         return ax

#     y = p[mask]
#     lo = intervals[:, 0]
#     hi = intervals[:, 1]

#     if ax is None:
#         fig, ax = plt.subplots()
#     for xi0, xi1, yi in zip(lo, hi, y):
#         ax.plot([xi0, xi1], [yi, yi])  # horizontal segment
#         if show_points:
#             ax.plot([xi0, xi1], [yi, yi], marker="o", linestyle="")

#     ax.set_xlabel("x")
#     ax.set_ylabel("p (height)")
#     if title:
#         ax.set_title(title)
#     # Make a little padding around data
#     x_min = np.min(lo)
#     x_max = np.max(hi)
#     if x_min == x_max:
#         x_min -= 0.5
#         x_max += 0.5
#     ax.set_xlim(x_min, x_max)
#     # y padding
#     y_min = np.min(y)
#     y_max = np.max(y)
#     if y_min == y_max:
#         y_min -= 0.5
#         y_max += 0.5
#     ax.set_ylim(0, 1)

#     return ax


def area_metric_hint():
    pass


# * --------------------------- the developments below are between scalar and P-box only


def distance_to_ecdf_bound(x0, quantile):
    """Min horizontal distance from x0 to the ECDF defined by quantile."""
    xs = sorted(quantile)
    i = bisect_left(xs, x0)
    if i == 0:
        return abs(xs[0] - x0)
    if i == len(xs):
        return abs(x0 - xs[-1])
    # nearest of the two neighbors
    return min(abs(x0 - xs[i - 1]), abs(xs[i] - x0))


def closer_bound(x0, left_edge, right_edge):
    """Decide which ECDF bound is closer to x0.

    args:
        x0: a scalar point
        left_edge: samples from the left bound of the ECDF
        x_right_samples: samples from the right bound of the ECDF
    """
    dl = distance_to_ecdf_bound(x0, left_edge)
    dr = distance_to_ecdf_bound(x0, right_edge)
    if dl < dr:
        return "left", dl, dr
    if dr < dl:
        return "right", dl, dr
    return "tie", dl, dr  # exactly equidistant


def if_outside(x0, left_edge, right_edge):
    """Check if x0 is outside the ECDF defined by left_edge and right_edge."""
    return x0 < left_edge[0] or x0 > right_edge[-1]


def if_right_in(x0, left_edge, right_edge):
    """Check if x0 is inside the ECDF defined by left_edge and right_edge."""
    return left_edge[-1] <= x0 <= right_edge[0]


# high-level func
def directional(x0, left_edge, right_edge):
    """give instructions on which direction to move the Pbox towards the scalar

    returns:
        output a message variable
    """
    a = if_outside(x0, left_edge, right_edge)  # a is boolean
    msg, _, _ = closer_bound(x0, left_edge, right_edge)

    if a:  # outside
        if msg == "left" or msg == "tie":
            return "out_left"
        elif msg == "right":
            return "out_right"
    else:  # inside
        if msg == "left" or msg == "tie":
            return "in_left"
        elif msg == "right":
            return "in_right"


def calibration_distance(a: Pbox, b: Number) -> float:
    """Estimate the calibration distance to compensate area metric between a P-box aand a scalar"""

    if not isinstance(a, Pbox):
        a, b = b, a  # swap so that a is always the Pbox

    msg, _, _ = closer_bound(b, a.left, a.right)

    if msg == "left" or msg == "tie":
        return np.abs(a.left[-1] - b)
    elif msg == "right":
        return np.abs(a.right[0] - b)


def slide_pbox_towards_scalar(a, b):
    """Slide the Pbox a towards the scalar b by one step.

    args:
        a: a Pbox
        b: a scalar

    returns:
        a new Pbox that is slid towards b by one step
    """
    # which direction to slide
    msg = directional(b, a.left, a.right)

    # how much to slide
    proposal_dd = calibration_distance(a, b)

    # worked and backup
    # match msg:
    #     case "out_right":
    #         # expand on the right
    #         return Staircase(a.left, a.right + proposal_dd)
    #     case "out_left":
    #         # expand on the left
    #         return Staircase(a.left - proposal_dd, a.right)
    #     case "in_left":
    #         # slide to the left
    #         return Staircase(a.left - proposal_dd, a.right - proposal_dd)
    #     case "in_right":
    #         # slide to the right
    #         return Staircase(a.left + proposal_dd, a.right + proposal_dd)

    match msg:
        case "out_right":
            # expand on the right
            d1 = 0
            d2 = proposal_dd
        case "out_left":
            # expand on the left
            d1 = proposal_dd
            d2 = 0
        case "in_left":
            # slide to the left
            d1 = proposal_dd
            d2 = -proposal_dd
        case "in_right":
            # slide to the right
            d1 = -proposal_dd
            d2 = proposal_dd

    return Staircase(a.left - d1, a.right + d2)



"""4 levels of epistemic propagation tests {low, b22, medium, high}"""

import pytest
import pyuncertainnumber as pun
from pyuncertainnumber import b2b, pba
from pyuncertainnumber.propagation.performance import foo_universal

""" Note

Some tests are commented out for automation but can run local tests """

# * ------------------------------ b2b level


@pytest.fixture
def example_intervals():
    a = pba.I(1, 5)
    b = pba.I(7, 13)
    c = pba.I(5, 10)
    return a, b, c


def test_b2b_level_direct_strategy(example_intervals):
    a, b, c = example_intervals

    y0 = b2b(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="direct",
    )
    assert y0 == pba.I(13.0, 148.0)


def test_b2b_level_endpoints_strategy(example_intervals):
    a, b, c = example_intervals

    y0 = b2b(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="endpoints",
    )
    assert y0 == pba.I(13.0, 148.0)


def test_b2b_level_subinterval_strategy(example_intervals):
    a, b, c = example_intervals

    y00 = b2b(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="subinterval",
        subinterval_style="direct",
        n_sub=10,
    )

    y22 = b2b(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="subinterval",
        n_sub=10,
        subinterval_style="endpoints",
    )

    assert y00 == pba.I(13.0, 148.0)
    assert y22 == pba.I(13.0, 148.0)


# def test_b2b_level_ga_opt_strategy(example_intervals):
#     a, b, c = example_intervals
#     # optimisation - ga
#     y3 = b2b(
#         vars=[a, b, c],
#         func=foo_universal,
#         interval_strategy="ga",
#     )

#     assert 12 <= y3.lo and y3.hi <= 149


# def test_b2b_level_ba_opt_strategy(example_intervals):
#     a, b, c = example_intervals

#     y3 = b2b(
#         vars=[a, b, c],
#         func=foo_universal,
#         interval_strategy="bo",
#     )

#     assert 12 <= y3.lo and y3.hi <= 149


# def test_b2b_level_cauchy_deviate_strategy(example_intervals):
#     a, b, c = example_intervals

#     y5 = b2b(
#         vars=[a, b, c],
#         func=foo_universal,
#         interval_strategy="cauchy_deviate",
#         n_sam=1000,
#     )


# * ------------------------------ high level


def test_high_level_endpoints_strategy():
    a = pun.I(1, 5)
    b = pun.I(7, 13)
    c = pun.I(5, 10)

    p = pun.Propagation(
        vars=[a, b, c],
        func=foo_universal,
        method="endpoints",
    )

    u = p.run()
    assert u.construct.lo == 13.0
    assert u.construct.hi == 148.0



"""
Test file for `tmcmc` implementation on a simple example of 2DOF system
"""

from numpy.typing import ArrayLike
import numpy as np
from pyuncertainnumber.calibration import pdfs
from pyuncertainnumber.calibration.tmcmc import TMCMC
from pyuncertainnumber import pba

# measurement data:
# eigen values of first mode
data1 = np.array([0.3860, 0.3922, 0.4157, 0.3592, 0.3615])
# eigen values of second mode
data2 = np.array([2.3614, 2.5877, 2.7070, 2.3875, 2.7272])
# eigen vector of first mode
data3 = np.array([1.68245252, 1.71103903, 1.57876073, 1.58722342, 1.61878479])

# number of particles (to approximate the posterior)
N = 100

# prior distribution of parameters
k1 = pdfs.Uniform(lower=0.8, upper=2.2)
k2 = pdfs.Uniform(lower=0.4, upper=1.2)

# Required! a list of all parameter objects
all_pars = [k1, k2]

k1 = pba.Distribution("uniform", (0.8, 2.2))
k2 = pba.Distribution("uniform", (0.4, 1.2))

# Required! a list of all parameter objects
all_pars = [k1, k2]


def log_likelihood_case3(particle_num: int, s: ArrayLike) -> float:
    """
    Required!

    log-likelihood function which is problem specific
    for the 2DOF example log-likelihood is

    args
        particle_num (int): int, particle number.

        s (ArrayLike) : numpy array of size Nop (number of parameters in all_pars)
            parameter vector of one particle in parameter space of dimension Nop, i.e. the particle's coordinates in parameter space.

    returns
        LL (float): log-likelihood function value.

    """
    sig1 = 0.0191
    sig2 = 0.0809  # = 0.05*1.618
    lambda1_s = (s[0] / 2 + s[1]) - np.sqrt(((s[0] / 2 + s[1]) ** 2 - s[0] * s[1]))
    phi12_s = (s[0] + s[1] - lambda1_s) / s[1]

    # see slide 21 -- case 3 of lecture notes
    LL = (
        np.log((2 * np.pi * sig1 * sig2) ** -5)
        + (-0.5 * (sig1 ** (-2)) * sum((lambda1_s - data1) ** 2))
        + (-0.5 * (sig2**-2) * sum((phi12_s - data3) ** 2))
    )
    return LL


def log_likelihood_case2(particle_num, s):
    """
    Log-likelihood for the 2DOF example - CASE 2 (two eigenvalues λ1 and λ2).

    Args:
        particle_num (int):
            Index of the particle (not used in this function, but required by the TMCMC framework).

        s (ArrayLike): numpy array of shape (2,)
            Parameter vector in this case:
                s[0] = q1 (stiffness parameter 1)
                s[1] = q2 (stiffness parameter 2)

    returns:
        LL (float): Log-likelihood value at this parameter vector.
    """
    q1 = s[0]
    q2 = s[1]

    # standard deviations (5% noise on true eigenvalues λ1=0.382, λ2=2.618)
    sig1 = 0.0191  # 0.05 * 0.382
    sig2 = 0.1309  # 0.05 * 2.618

    # compute eigenvalues λ1(q1, q2) and λ2(q1, q2) for the 2×2 system
    # characteristic equation: λ^2 - (q1 + 2 q2) λ + q1 q2 = 0
    # closed-form solution:
    # λ₁,₂ = (q1/2 + q2) ∓ sqrt((q1/2 + q2)^2 - q1 q2)
    center = q1 / 2.0 + q2
    disc = center**2 - q1 * q2
    if disc < 0:
        # if the discriminant is negative, eigenvalues are complex -> impossible here physically
        # give a very low likelihood
        return -np.inf

    sqrt_disc = np.sqrt(disc)
    lambda1_s = center - sqrt_disc
    lambda2_s = center + sqrt_disc

    # Gaussian likelihood for 5 measurements of λ1 and 5 of λ2
    # p(d | θ) ∝ exp( -1/(2σ1²) Σ (λ1_s - d1_m)² - 1/(2σ2²) Σ (λ2_s - d2_m)² )
    # log p(d | θ) = const - 0.5/σ1² Σ (λ1_s - d1_m)² - 0.5/σ2² Σ (λ2_s - d2_m)²

    # constant term (same form as in your Case 3 implementation)
    const_term = np.log((2 * np.pi * sig1 * sig2) ** -5)

    # misfit terms
    misfit1 = -0.5 * (sig1**-2) * np.sum((lambda1_s - data1) ** 2)
    misfit2 = -0.5 * (sig2**-2) * np.sum((lambda2_s - data2) ** 2)

    LL = const_term + misfit1 + misfit2
    return LL


def test_2dof_tmcmc(tmp_path):
    """Test TMCMC on 2DOF example"""
    # Use temporary directory for status file
    status_file = tmp_path / "status_file_2DOF_case2_pba.txt"

    t = TMCMC(
        N=N,
        parameters=all_pars,
        names=["theta_1", "theta_2"],
        log_likelihood=log_likelihood_case2,
        status_file_name=str(status_file),
    )

    mytrace = t.run()

    assert mytrace is not None



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
import scipy.stats as sps
import matplotlib.pyplot as plt
from warnings import *
from dataclasses import dataclass, field
from ..characterisation.utils import pl_pcdf, pl_ecdf
from .params import Params
from .pbox_parametric import named_pbox
import statsmodels.distributions.copula as Copula
from .dependency import Dependency
from statsmodels.distributions.copula.api import CopulaDistribution
from .pbox_abc import Staircase
from .ecdf import get_ecdf
from numbers import Number
from .mixins import NominalValueMixin
from numpy.typing import ArrayLike


if TYPE_CHECKING:
    from pyuncertainnumber import Interval

"""distribution constructs"""


# * --------------------- parametric cases --------------------- *#


@dataclass
class Distribution(NominalValueMixin):
    """Two signatures are currentlly supported, either a parametric specification or from a nonparametric empirical data set.

    note:
        the nonparametric instasntiation via arrtribute `empirical_data` will be deprecated soon.
        We have introduced a :class:`distributions.ECDF` class instead.

    example:
        >>> d = Distribution('gaussian', (0,1))
    """

    dist_family: str = None
    dist_params: list[float] | tuple[float, ...] = None
    empirical_data: list[float] | np.ndarray = None
    skip_post: bool = False

    def __post_init__(self):
        if self.skip_post:
            return
        elif all(
            v is None for v in [self.dist_family, self.dist_params, self.empirical_data]
        ):
            raise ValueError(
                "At least one of dist_family, dist_params or sample must be specified"
            )
        self.dist = self.rep()

    def __repr__(self):
        # if self.empirical_data is not None:
        #     return "sample-approximated distribution object"
        if self.dist_params is not None:
            return f"dist ~ {self.dist_family}{self.dist_params}"
        elif self.empirical_data is not None:
            return "dist ~ sample-approximated distribution object"
        else:
            return "blank object"

    def rep(self):
        """Create the underling dist object either sps dist or sample approximated or pbox dist

        note:
            underlying constructor to create the scipy.stats distribution object
        """
        if self.dist_family is not None:
            return self._match_distribution()

    #! **kwargs from sps.dist will yield error
    def _match_distribution(self):
        """match the distribution object based on the family and parameters"""
        params = self.dist_params
        if not isinstance(params, (tuple, list)):
            params = (params,)

        # TODO: special cases make it general
        if self.dist_family == "lognormal":  # special case for lognormal distribution
            return lognormal_sane(*params)
        return named_dists.get(self.dist_family)(*params)

    def parse_params_from_dist(self):
        self.dist_params = tuple(self._dist.args) + tuple(
            v for k, v in sorted(self._dist.kwds.items())
        )

    def flag(self):
        """boolean flag for if the distribution is a parameterised distribution or not
        note:
            - only parameterised dist can do sampling
            - for non-parameterised sample-data based dist, next steps could be fitting
        """
        if (self.dist_params is not None) & (self.dist_family is not None):
            self._flag = True
        else:
            self._flag = False

    def sample(self, size):
        """generate deviates from the distribution"""
        if self._flag:
            return self._dist.rvs(size=size)
        else:
            raise ValueError(
                "Sampling not supported for sample-approximated distributions"
            )

    def generate_rns(self, N):
        """generate 'N' random numbers from the distribution"""
        return self.sample(N)

    def alpha_cut(self, alpha):
        """alpha cut interface"""
        return self._dist.ppf(alpha)

    def make_nominal_value(self):
        """one value representation of the distribution
        note:
            - use mean for now;
        """
        if self._flag:
            self._naked_value = self._dist.mean()
        else:
            self._naked_value = np.mean(self.empirical_data)

    def plot(self, **kwargs):
        """display the distribution"""
        if self.empirical_data is not None:
            return pl_ecdf(self.empirical_data, **kwargs)
        pl_pcdf(self._dist, **kwargs)

    def display(self, **kwargs):
        self.plot(**kwargs)
        plt.show()

    def _get_hint(self):
        pass

    def fit(self, data):
        """fit the distribution to the data"""
        pass

    def get_PI(self, alpha: Number = 0.95) -> Interval:
        """Compute the predictive interval at the coverage level of `alpha`

        args:
            - alpha (float): coverage level, default is 0.95

        example:
            >>> from pyuncertainnumber import pba
            >>> d = pba.Distribution('gaussian', (0, 1))
            >>> pi = d.get_PI(alpha=0.95)
            >>> print(pi)  # prints the interval at the 95% coverage level
        """

        from pyuncertainnumber import Interval

        lo_cut_level = (1 - alpha) / 2  # 0.025
        hi_cut_level = 1 - lo_cut_level  # 0.975

        hi = self.alpha_cut(hi_cut_level)
        lo = self.alpha_cut(lo_cut_level)
        return Interval(lo, hi)

    def pdf(self, x: ArrayLike):
        """compute the probability density function (pdf) at x"""
        return self._dist.pdf(x)

    def log_pdf_eval(self, x: ArrayLike):
        """compute the log of probability density function (pdf) at x"""
        return self._dist.logpdf(x)

    def cdf(self, x: ArrayLike):
        """compute the cumulative distribution function (cdf) at x"""
        return self._dist.cdf(x)

    def get_whole_cdf(self):
        """return the cumulative distribution function (cdf)"""
        return self._dist.cdf(Params.p_values)

    def _compute_nominal_value(self):
        return np.round(self._naked_value, 3)

    @property
    def dist(self):
        """the underlying sps.dist object"""
        return self._dist

    @dist.setter
    def dist(self, value):
        self._dist = value
        if self.dist_params is None:
            self.parse_params_from_dist()
        self.flag()
        self.make_nominal_value()

    @property
    def lo(self):
        return self._dist.ppf(Params.p_lboundary)

    @property
    def hi(self):
        return self._dist.ppf(Params.p_hboundary)

    @property
    def range(self):
        from pyuncertainnumber import Interval

        return Interval(self.lo, self.hi)

    @property
    def hint(self):
        pass

    # *  ---------------------constructors---------------------* #
    @classmethod
    def dist_from_sps(
        cls, dist: sps.rv_continuous | sps.rv_discrete, shape: str = None
    ):
        # old version but does not work with 'scales'
        # params = dist.args + tuple(dist.kwds.values())
        # return cls(dist_family=shape, dist_params=params)
        # params = {"args": dist.args, "kwds": dist.kwds}
        # return named_dists.get(shape)(*params["args"], **params["kwds"])
        obj = cls(
            dist_family=shape, dist_params=None, empirical_data=None, skip_post=True
        )
        obj.dist = dist
        return obj

    # @classmethod
    # def dist_from_sps(
    #     cls, dist: sps.rv_continuous | sps.rv_discrete, shape: str = None
    # ):
    #     obj = cls.__new__(cls)  # bypass __init__ + __post_init__ logic
    #     obj.__dict__.update(
    #         dist_family=shape,
    #         dist_params=None,
    #         empirical_data=None,
    #         _dist=dist,
    #         _skip_post=True,
    #     )
    #     return obj

    # *  ---------------------conversion---------------------* #

    def to_pbox(self):
        """convert the distribution to a pbox
        note:
            - this only works for parameteried distributions for now
            - later on work with sample-approximated dist until `fit()`is implemented
        """
        if self._flag:
            params = self.dist_params
            if not isinstance(params, (tuple, list)):
                params = (params,)
            return named_pbox.get(self.dist_family)(*params)

    def __neg__(self):
        return -self.to_pbox()

    def __add__(self, other):
        p = self.to_pbox()
        return p.add(other, dependency="f")

    def __radd__(self, other):
        return self.add(other, dependency="f")

    def __sub__(self, other):
        p = self.to_pbox()
        return p.sub(other, dependency="f")

    def __rsub__(self, other):
        self = -self
        return self.add(other, dependency="f")

    def __mul__(self, other):
        p = self.to_pbox()
        return p.mul(other, dependency="f")

    def __rmul__(self, other):
        return self.mul(other, dependency="f")

    def __truediv__(self, other):
        p = self.to_pbox()
        return p.div(other, dependency="f")

    def __rtruediv__(self, other):
        p = self.to_pbox()
        try:
            return other * p.recip()
        except:
            return NotImplemented

    def __pow__(self, other):
        p = self.to_pbox()
        return p.pow(other, dependency="f")

    def __rpow__(self, other):
        if not hasattr(other, "__iter__"):
            other = np.array((other))
        p = self.to_pbox()
        return p.pow(other, dependency="f")


class JointDistribution:
    """Joint distribution class


    example:
        >>> from pyuncertainnumber import pba
        >>> dist_a, dist_b = pba.Distribution('gaussian', (5,1)), pba.Distribution('uniform', (2, 3))
        >>> c = pba.Dependency('gaussian', params=0.8)
        >>> joint_dist = pba.JointDistribution(copula=c, marginals=[dist_a, dist_b])
        >>> samples = joint_dist.sample(size=1000)
    """

    def __init__(
        self,
        copula: Dependency,
        marginals: list[Distribution],
    ):
        self.marginals = marginals
        self.copula = copula
        self._joint_dist = CopulaDistribution(
            copula=self.copula._copula, marginals=[m._dist for m in self.marginals]
        )
        self.ndim = len(self.marginals)

    def __repr__(self):
        return f"a {self.ndim}-dimensional JointDistribution with copula: {self.copula.family} and marginals: {[m.dist_family for m in self.marginals]}"

    @staticmethod
    def from_sps(copula: Copula, marginals: list[sps.rv_continuous]):
        return CopulaDistribution(copula=copula, marginals=marginals)

    def sample(self, size, random_state=42):
        """generate orginal-space samples from the joint distribution"""
        return self._joint_dist.rvs(size, random_state=random_state)

    def u_sample(self, size, random_state=42):
        """generate copula-space samples from the joint distribution"""
        return self.copula.rvs(size, random_state=random_state)

    def joint_density_of_bi_grid(self, x: ArrayLike, y: ArrayLike):
        """compute the joint density on a grid given x and y arrays

        Used for bivariate arithmetic calculations of X and Y with designated (known) dependency and marginals.


        note:
            discretisation step sizes dx and dy are set up by the input x and y arrays

        example:
            >>> x = np.linspace(0, 1, 50)
            >>> y = np.linspace(0, 1, 50)
            >>> dep = Dependency("gaussian", params=0.7)
            >>> joint_density = dep.joint_density_of_grid(x, y)
        """
        XX, YY = np.meshgrid(x, y, indexing="ij")
        UU, VV = self.marginals[0].cdf(XX), self.marginals[1].cdf(
            YY
        )  # transform to unit square
        uv = np.column_stack([UU.ravel(), VV.ravel()])
        c_uv = self.copula.pdf(uv).reshape(XX.shape)  # copula density
        fX = self.marginals[0].pdf(XX)
        fY = self.marginals[1].pdf(YY)
        fXY = c_uv * fX * fY
        dx = x[1] - x[0]
        dy = y[1] - y[0]
        return XX, YY, fXY, dx, dy

    @staticmethod
    def cdf_of_g(XX, YY, fXY, dx, dy, g_func, z_vals) -> ArrayLike:
        """Numerically approximate F_Z(z) = P(g(X,Y) <= z) via discretisation on a grid

        args:
            z_vals (ArrayLike): discretisation of z values (array) at which to compute F_Z
            XX, YY: the grid arrays from meshgrid
            fXY: joint density on the grid
            dx, dy: spacings in x and y directions
            g_func (callable): a general callable applied elementwise to (XX, YY)

        returns:
            FZ (ArrayLike): cumulative distribution function values at z_vals

        note:
            given precomputed grid (XX,YY), joint density fXY, and spacings.
        """
        G = g_func(XX, YY)  # evaluate g on the grid once
        FZ = np.empty_like(z_vals, dtype=float)
        for k, z in enumerate(z_vals):
            mask = G <= z
            FZ[k] = np.sum(fXY * mask) * dx * dy
        # Ensure numeric monotonicity (optional small fix)
        FZ = np.clip(np.maximum.accumulate(FZ), 0.0, 1.0)
        return FZ


# * --------------------- non-parametric ecdf cases --------------------- *#


class ECDF(Staircase):
    """Empirical cumulative distribution function (ecdf) class

    .. admonition:: Implementation

        supported by `Pbox` API hence samples will be degenerate intervals

    example:
        >>> import numpy as np
        >>> s = np.random.normal(size=1000)
        >>> ecdf = ECDF(s)
        >>> ecdf.plot()
    """

    def __init__(self, empirical_data: np.ndarray):
        left, p_values = get_ecdf(empirical_data)
        # TODO: quantile direct into Staircase. Hmm...
        super().__init__(left=left, right=left)


# * ------------------ special sane cases ------------------ *#
def uniform_sane(a, b):
    return sps.uniform(loc=a, scale=b - a)


def lognormal_sane(mu, sigma):
    """The sane lognormal which creates a lognormal distribution object based on the mean (mu) and standard deviation (sigma)
    of the underlying normal distribution.

    args:
        - mu (float): Mean of the underlying normal distribution
        - sigma (float): Standard deviation of the underlying normal distribution

    Returns:
        - A scipy.stats.lognorm frozen distribution object
    """
    shape = sigma  # shape parameter for lognorm
    scale = np.exp(mu)  # scale parameter is exp(mu)
    return sps.lognorm(s=shape, scale=scale)


class LognormalSaneDist:
    def ppf(self, p_values, *params):
        dist = lognormal_sane(*params)
        return dist.ppf(p_values)

    def stats(self, *params, moments="mv"):
        dist = lognormal_sane(*params)
        return dist.stats(moments=moments)


def expon_sane(lamb):
    """Sane exponential distribution constructor"""
    return sps.expon(scale=1 / lamb)


class WrapperDist:
    def __init__(self, ppf_func):
        self._ppf = ppf_func

    def ppf(self, *args, **kwargs):
        return self._ppf(*args, **kwargs)


named_dists = {
    "alpha": sps.alpha,
    "anglit": sps.anglit,
    "arcsine": sps.arcsine,
    "argus": sps.argus,
    "beta": sps.beta,
    "betaprime": sps.betaprime,
    "bradford": sps.bradford,
    "burr": sps.burr,
    "burr12": sps.burr12,
    "cauchy": sps.cauchy,
    "chi": sps.chi,
    "chi2": sps.chi2,
    "cosine": sps.cosine,
    "crystalball": sps.crystalball,
    "dgamma": sps.dgamma,
    "dweibull": sps.dweibull,
    "erlang": sps.erlang,
    "expon": sps.expon,
    "exponential": sps.expon,  # re-engineered exponential distribution
    "exponential_by_lambda": expon_sane,  # re-engineered exponential distribution
    "exponnorm": sps.exponnorm,
    "exponweib": sps.exponweib,
    "exponpow": sps.exponpow,
    "f": sps.f,
    "fatiguelife": sps.fatiguelife,
    "fisk": sps.fisk,
    "foldcauchy": sps.foldcauchy,
    "foldnorm": sps.foldnorm,
    # 'frechet_r' : sps.frechet_r,
    # 'frechet_l' : sps.frechet_l,
    "genlogistic": sps.genlogistic,
    "gennorm": sps.gennorm,
    "genpareto": sps.genpareto,
    "genexpon": sps.genexpon,
    "genextreme": sps.genextreme,
    "gausshyper": sps.gausshyper,
    "gamma": sps.gamma,
    "gengamma": sps.gengamma,
    "genhalflogistic": sps.genhalflogistic,
    "geninvgauss": sps.geninvgauss,
    # 'gibrat' : sps.gibrat,
    "gompertz": sps.gompertz,
    "gumbel_r": sps.gumbel_r,
    "gumbel_l": sps.gumbel_l,
    "halfcauchy": sps.halfcauchy,
    "halflogistic": sps.halflogistic,
    "halfnorm": sps.halfnorm,
    "halfgennorm": sps.halfgennorm,
    "hypsecant": sps.hypsecant,
    "invgamma": sps.invgamma,
    "invgauss": sps.invgauss,
    "invweibull": sps.invweibull,
    "johnsonsb": sps.johnsonsb,
    "johnsonsu": sps.johnsonsu,
    "kappa4": sps.kappa4,
    "kappa3": sps.kappa3,
    "ksone": sps.ksone,
    "kstwobign": sps.kstwobign,
    "laplace": sps.laplace,
    "levy": sps.levy,
    "levy_l": sps.levy_l,
    "levy_stable": sps.levy_stable,
    "logistic": sps.logistic,
    "loggamma": sps.loggamma,
    "loglaplace": sps.loglaplace,
    # "lognorm": sps.lognorm,
    "lognormal": LognormalSaneDist(),
    "loguniform": sps.loguniform,
    "lomax": sps.lomax,
    "maxwell": sps.maxwell,
    "mielke": sps.mielke,
    "moyal": sps.moyal,
    "nakagami": sps.nakagami,
    "ncx2": sps.ncx2,
    "ncf": sps.ncf,
    "nct": sps.nct,
    "norm": sps.norm,
    "normal": sps.norm,
    "gaussian": sps.norm,
    "norminvgauss": sps.norminvgauss,
    "pareto": sps.pareto,
    "pearson3": sps.pearson3,
    "powerlaw": sps.powerlaw,
    "powerlognorm": sps.powerlognorm,
    "powernorm": sps.powernorm,
    "rdist": sps.rdist,
    "rayleigh": sps.rayleigh,
    "rice": sps.rice,
    "recipinvgauss": sps.recipinvgauss,
    "semicircular": sps.semicircular,
    "skewnorm": sps.skewnorm,
    "t": sps.t,
    "trapz": sps.trapz,
    "triang": sps.triang,
    "truncexpon": sps.truncexpon,
    "truncnorm": sps.truncnorm,
    "tukeylambda": sps.tukeylambda,
    "uniform": uniform_sane,
    "vonmises": sps.vonmises,
    "vonmises_line": sps.vonmises_line,
    "wald": sps.wald,
    "weibull": sps.weibull_min,
    "weibull_min": sps.weibull_min,
    "weibull_max": sps.weibull_max,
    "wrapcauchy": sps.wrapcauchy,
    "bernoulli": sps.bernoulli,
    "betabinom": sps.betabinom,
    "binom": sps.binom,
    "binomial": sps.binom,
    "boltzmann": sps.boltzmann,
    "dlaplace": sps.dlaplace,
    "geom": sps.geom,
    "hypergeom": sps.hypergeom,
    "logser": sps.logser,
    "nbinom": sps.nbinom,
    "planck": sps.planck,
    "poisson": sps.poisson,
    "randint": sps.randint,
    "skellam": sps.skellam,
    "zipf": sps.zipf,
    "yulesimon": sps.yulesimon,
}



from functools import partial
from abc import ABC, abstractmethod
from pyuncertainnumber import pba
from ..characterisation.uncertainNumber import UncertainNumber
from ..pba.pbox_abc import Pbox
from ..pba.intervals.number import Interval
from ..pba.distributions import Distribution
from ..pba.dependency import Dependency
from .b2b import b2b
from ..decorator import constructUN
from .local_optimisation import local_optimisation_method
from .mixed_up import (
    interval_monte_carlo,
    slicing,
    double_monte_carlo,
)
from ..pba.intervals.intervalOperators import make_vec_interval
import numpy as np


"""the new top-level module for the propagation of uncertain numbers"""

import logging

logging.basicConfig(level=logging.INFO)


# * ------------------ constructs Propagation ------------------ *
class P(ABC):
    """Base class blueprint. Not for direct use"""

    def __init__(self, vars, func, method, dependency=None):
        self._vars = vars
        self.func = func
        self.method = method
        self.dependency = dependency

    def post_init_check(self):
        """some checks"""

        assert callable(self.func), "function is not callable"
        self.type_check()
        self.method_check()

    @abstractmethod
    def type_check(self):
        """if the nature of the UN suitable for the method"""
        pass

    @abstractmethod
    def method_check(self):
        """if the method is suitable for the nature of the UN"""
        pass


class AleatoryPropagation(P):
    """Aleatoric uncertainty propagation class for Distribution constructs only

    args:
        vars (Distribution): a list of uncertain numbers objects

        func (callable): the response or performance function applied to the uncertain numbers

        method (str): a string indicating the method to be used for propagation.

        dependency (string or Dependency): a Dependency object(i.e. a copula function) to model the dependency structure among input variables.
            Strings such as "independence" accepted for independence.


    note:
        Supported methods include "monte_carlo". Note that "taylor_expansion" is not supported herein but implemented as a standalone function in the module `taylor_expansion.py`.

    caution:
        This function supports with low-level constructs NOT the high-level `UN` (uncertain number) objects.
        For `UN` objects, use `Propagation` class as an high-level API.


        .. seealso::
            :func:`Propagation` : the high-level API for uncertain number propagation.


    example:
        >>> from pyuncertainnumber import pba
        >>> from pyuncertainnumber.propagation.p import AleatoryPropagation
        >>> def foo(x): return x[0] ** 3 + x[1] + x[2]
        >>> a_d = pba.Distribution('gaussian', (3,1))
        >>> b_d = pba.Distribution('gaussian', (10, 1))
        >>> c_d = pba.Distribution('uniform', (5, 10))
        >>> aleatory = AleatoryPropagation(vars=[a_d, b_d, c_d], func=foo, method='monte_carlo')
        >>> result = aleatory(n_sam=1000)
    """

    from .taylor_expansion import taylor_expansion_method

    def __init__(self, vars, func, method, dependency=None):
        super().__init__(vars, func, method, dependency)
        self.post_init_check()

    def type_check(self):
        """only distributions"""
        from ..pba.distributions import Distribution
        from ..pba.pbox_abc import Pbox

        assert all(
            isinstance(v, Distribution | Pbox) for v in self._vars
        ), "Not all variables are distributions"

    def method_check(self):
        assert self.method in [
            "monte_carlo",
            "latin_hypercube",
        ], "Method not supported for aleatory uncertainty propagation"

    def run(self, n_sam: int = 1000):
        """doing the propagation"""
        match self.method:
            case "monte_carlo":
                if self.dependency is None or self.dependency == "independence":
                    # regular sampling style
                    try:
                        # regular sampling style
                        input_samples = [v.sample(n_sam) for v in self._vars]
                        output_samples = self.func(input_samples)
                    except Exception as e:
                        # vectorised sampling style
                        input_samples = np.array(
                            [v.sample(n_sam) for v in self._vars]
                        ).T  # (n_sam, n_vars) == (n, d)
                        output_samples = self.func(input_samples)
                        return output_samples
                else:
                    j = pba.JointDistribution(
                        copula=self.dependency, marginals=self._vars
                    )

                    s_inputs = j.sample(n_sam)  # (n_sam, n_vars)

                    output_samples = self.func(s_inputs)
                    return output_samples
            case _:
                raise ValueError("method not yet supported")
        return output_samples


class EpistemicPropagation(P):
    """Epistemic uncertainty propagation class for construct

    args:
        vars (Interval): a list of interval objects

        func (callable): the response or performance function applied to the uncertain numbers

        method (str): a string indicating the method to be used for propagation

        interval_strategy (str): a strategy for interval propagation, including {'endpoints', 'subinterval'}

    caution:
        This function supports with low-level constructs NOT the high-level `UN` (uncertain number) objects.
        For `UN` objects, use `Propagation` class as an high-level API.


        .. seealso::
            :func:`Propagation` : the high-level API for uncertain number propagation.

    example:
        >>> from pyuncertainnumber import pba
        >>> from pyuncertainnumber.propagation.p import EpistemicPropagation
        >>> def foo(x): return x[0] ** 3 + x[1] + x[2]
        >>> a = pba.I(1, 5)
        >>> b = pba.I(7, 13)
        >>> c = pba.I(5, 10)
        >>> ep = EpistemicPropagation(vars=[a,b,c], func=foo, method='subinterval')
        >>> result = ep(n_sub=20, style='endpoints')
    """

    def __init__(self, vars, func, method):
        super().__init__(vars, func, method)
        self.post_init_check()

    def type_check(self):
        """only intervals"""

        from ..pba.intervals.number import Interval

        assert all(
            isinstance(v, Interval) for v in self._vars
        ), "Not all variables are intervals"

    def method_check(self):
        pass

    def run(self, **kwargs):
        #! caveat: possibly requires more kwargs for some methods
        """doing the propagation"""
        match self.method:
            case "endpoint" | "endpoints" | "vertex":
                handler = partial(b2b, interval_strategy="endpoints")
            case "subinterval" | "subintervals" | "subinterval_reconstitution":
                handler = partial(b2b, interval_strategy="subinterval")
            case (
                "cauchy"
                | "cauchy_deviate_method"
                | "endpoint_cauchy"
                | "endpoints_cauchy"
            ):
                handler = partial(b2b, interval_strategy="cauchy_deviate")
            case (
                "local_optimization"
                | "local_optimisation"
                | "local optimisation"
                | "local optimization"
            ):
                handler = local_optimisation_method
            case (
                "genetic_optimisation"
                | "genetic_optimization"
                | "genetic optimization"
                | "genetic optimisation"
            ):
                handler = partial(b2b, interval_strategy="ga")
            case "bayesian_optimisation" | "bo":
                handler = partial(b2b, interval_strategy="bo")
            case _:
                raise ValueError("Unknown method")

        results = handler(
            make_vec_interval(self._vars),  # pass down vec interval
            self.func,
            **kwargs,
        )
        return results


class MixedPropagation(P):
    """Mixed uncertainty propagation class for construct

    args:
        vars (Pbox or DempsterShafer): a list of uncertain numbers objects

        func (callable): the response or performance function applied to the uncertain numbers

        method (str): a string indicating the method to be used for pbox propagation, including {'interval_monte_carlo', 'slicing', 'double_monte_carlo'}.

        dependency (string or Dependency): a Dependency object(i.e. a copula function) to model the dependency structure among input variables.
            Strings such as "independence" accepted for independence.

        interval_strategy (str): a sub-level strategy selector for interval propagation, including {'direct', 'subinterval', 'endpoints'}.

    caution:
        This function supports with low-level constructs NOT the high-level `UN` (uncertain number) objects.
        For `UN` objects, use `Propagation` class as an high-level API.


        .. seealso::
            :func:`Propagation` : the high-level API for uncertain number propagation.


    warning:
        The computation cost increases exponentially with the number of input variables and the number of slices.
        Be cautious with the choice of number of slices ``n_slices`` given the number of input variables ``vars`` of the response function.


    note:
        Discussion of the methods and strategies. When choosing ``interval_strategy``, "direct" requires function signature to take a list of inputs,
        whereas "subinterval" and "endpoints" require the function to take a vectorised signature. Currently, only "interval_monte_carlo" supports with dependency structures (e.g. copulas).

        When calling the `run` function to do propagation, extra keyword arguments are needed to be passed down to the selected `method`.
        For example, `n_sam` for "interval_monte_carlo"; `n_slices` for "slicing"; `n_outer`, `n_inner` for "double_monte_carlo".

    example:
        >>> from pyuncertainnumber import pba
        >>> from pyuncertainnumber.propagation.p import MixedPropagation
        >>> def foo(x): return x[0] ** 3 + x[1] + x[2]
        >>> a = pba.normal([2, 3], [1])
        >>> b = pba.normal([10, 14], [1])
        >>> c = pba.normal([4, 5], [1])
        >>> mix = MixedPropagation(vars=[a,b,c], func=foo, method='slicing', interval_strategy='subinterval')
        >>> result = mix.run(n_slices=20, n_sub=2, style='endpoints')
    """

    def __init__(self, vars, func, method, dependency=None, interval_strategy=None):

        super().__init__(vars, func, method, dependency)
        self.interval_strategy = interval_strategy
        self.post_init_check()

    # assume striped UN classes (i.e. constructs only)
    def type_check(self):
        """Inspection if inputs are mixed uncertainy model"""

        has_I = any(isinstance(item, Interval) for item in self._vars)
        has_D = any(isinstance(item, Distribution) for item in self._vars)
        has_P = any(isinstance(item, Pbox) for item in self._vars)

        assert (has_I and has_D) or has_P, "Not a mixed uncertainty problem"

    def method_check(self):
        """Check if the method is suitable for mixed uncertainty propagation"""
        assert self.method in [
            "interval_monte_carlo",
            "slicing",
            "double_monte_carlo",
        ], f"Method {self.method} not supported for mixed uncertainty propagation"

    def run(self, **kwargs):
        """doing the propagation. Extra keyword are needed to be passed down to the selected `method`"""
        match self.method:
            case "interval_monte_carlo":
                if self.dependency is None or self.dependency == "independence":
                    handler = partial(interval_monte_carlo, dependency=None)
                else:
                    imc_w_d = partial(interval_monte_carlo, dependency=self.dependency)
                    handler = imc_w_d
            case "slicing":
                handler = slicing
            case "double_monte_carlo":
                handler = double_monte_carlo
            case None:
                handler = slicing
            case _:
                raise ValueError("Unknown method")

        results = handler(self._vars, self.func, self.interval_strategy, **kwargs)
        return results


# * ------------------ Uncertain Number Propagation ------------------ *
class Propagation:
    """High-level integrated class for the propagation of uncertain numbers

    args:
        vars (UncertainNumber): a list of uncertain numbers objects

        func (Callable): the response or performance function applied to the uncertain numbers

        method (str):
            a string indicating the method to be used for propagation (e.g. "monte_carlo", "endpoint", etc.) which may depend on the constructs of the uncertain numbers.
            See notes about function signature.

        dependency (string or Dependency): a Dependency object(i.e. a copula function) to model the dependency structure among input variables.
            Strings such as "independence" accepted for independence.

        interval_strategy (str):
            a strategy for interval propagation, including {'direct', 'subinterval', 'endpoints'} which will
            affect the function signature of the response function. See notes about function signature.

    caution:
        This class supports with high-level computation with `UncertainNumber` objects.

    note:
        Discussion of the methods and strategies.
        When choosing ``interval_strategy``, "direct" requires function signature to take a list of inputs,
        whereas "subinterval" and "endpoints" require the function to take a vectorised signature.

    warning:
        The computation cost increases exponentially with the number of input variables and the number of slices.
        Be cautious with the choice of number of slices ``n_slices`` given the number of input variables ``vars`` of the response function.

    example:
        >>> import pyuncertainnumber as pun
        >>> # construction of uncertain number objects
        >>> a = pun.I(2, 3)
        >>> b = pun.normal(4, 1)
        >>> c = pun.uniform([4,5], [9,10])

        >>> # vectorised function signature with matrix input (2D np.ndarray)
        >>> def foo_vec(x): return x[:, 0] ** 3 + x[:, 1] + x[:, 2]

        >>> # high-level propagation API
        >>> p = Propagation(vars=[a,b,c],
        >>>     func=foo,
        >>>     method='slicing',
        >>>     interval_strategy='subinterval'
        >>> )

        >>> # heavy-lifting of propagation
        >>> t = p.run(n_sam=20, n_sub=2, style='endpoints')
    """

    def __init__(
        self,
        vars: list[UncertainNumber],
        func: callable,
        method: str,
        dependency: str | Dependency = None,
        interval_strategy: str = None,
    ):

        self._vars = vars
        self._func = func
        self.method = method
        self.dependency = dependency
        self.interval_strategy = interval_strategy
        self._post_init_check()

    def _post_init_check(self):
        """Some checks after initialisation"""

        # strip the underlying constructs from UN
        self._constructs = [c._construct for c in self._vars]

        # supported methods check

        # assign method herein
        self.assign_method()

    def assign_method(self):
        """Assign the propagation method based on the essence of constructs"""
        # created an underlying propagation `self.p` object

        # all
        all_I = all(isinstance(item, Interval) for item in self._constructs)
        all_D = all(isinstance(item, Distribution) for item in self._constructs)
        # any
        has_I = any(isinstance(item, Interval) for item in self._constructs)
        has_D = any(isinstance(item, Distribution) for item in self._constructs)
        has_P = any(isinstance(item, Pbox) for item in self._constructs)

        if all_I:
            # all intervals
            logging.info("interval propagation")
            self.p = EpistemicPropagation(self._constructs, self._func, self.method)
        elif all_D:
            logging.info("distribution propagation")
            # all distributions
            self.p = AleatoryPropagation(
                self._constructs, self._func, self.method, self.dependency
            )
        elif (has_I and has_D) or has_P:
            # mixed uncertainty
            logging.info("mixed uncertainty propagation")
            self.p = MixedPropagation(
                vars=self._constructs,
                func=self._func,
                method=self.method,
                dependency=self.dependency,
                interval_strategy=self.interval_strategy,
                # interval_strategy=self.kwargs.get("interval_strategy", None),
            )
        else:
            raise ValueError(
                "Not a valid combination of uncertainty types. "
                "Please check the input variables."
            )

    @property
    def constructs(self):
        """return the underlying constructs"""
        return self._constructs

    @constructUN
    def run(self, **kwargs):
        """Doing the propagation and return UN

        return:
            UncertainNumber: the result of the propagation as an uncertain number object
        """

        # choose the method accordingly
        return self.p.run(**kwargs)
